/*****************************************************************************
 *
 * FILE:	html32.c
 * DESCRIPTION:	Basilisk: HTML3.2 printer
 * DATE:	Fri, Mar  8 2002
 * UPDATE:	Sun, Aug 19 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2002-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2002-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: html32.c,v 1.4 2007/08/19 16:48:36 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include "basilisk.h"
#include "functions.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	html	ver.HTML32

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	html32Done(void);
static void	htmlStart(const String);
static void	htmlEnd(void);
static void	bodyStart(Attrib *);
static void	bodyEnd(void);
static void	h1(const String, halign_t);
static void	h2(const String, halign_t);
static void	h3(const String, halign_t);
static void	h4(const String, halign_t);
static void	h5(const String, halign_t);
static void	h6(const String, halign_t);
static void	headings(int, const String, halign_t);
static void	addressBegin(void);
static void	addressEnd(void);
static void	preBegin(void);
static void	preEnd(void);
static void	centerBegin(void);
static void	centerEnd(void);
static void	pBegin(halign_t);
static void	pEnd(void);
static void	divBegin(halign_t);
static void	divEnd(void);
static void	block_begin(const String, halign_t);
static void	block_end(const String);
static void	ulBegin(Attrib *);
static void	ulEnd(void);
static void	olBegin(Attrib *);
static void	olEnd(void);
static void	li(const String);
static void	aHref(const String, const String);
static void	aName(const String, const String);
static void	strong(const String);
static void	font(const String, const String, const String);
static void	br(void);
static void	hr(Attrib *);
static void	img(const String, const String,
		    Pixels, Pixels, Pixels, align_t);
static void	formStart(const String, method_t);
static void	multipartStart(const String, method_t);
static void	formEnd(void);
static void	inputText(const String, const String,
			  unsigned int, unsigned int);
static void	inputPassword(const String, const String,
			      unsigned int, unsigned int);
static void	inputCheckbox(const String, const String, bool);
static void	inputRadio(const String, const String, bool);
static void	inputSubmit(const String, const String);
static void	inputReset(const String, const String);
static void	inputHidden(const String, const String);
static void	inputImage(const String, const String, align_t);
static void	inputFile(const String, const String,
			  unsigned int, unsigned int);
static void	selectStart(const String, unsigned int, bool);
static void	selectEnd(void);
static void	selectOption(const String, bool);
static void	textareaStart(const String, unsigned int, unsigned int);
static void	textareaEnd(void);
static void	tableStart(Attrib *);
static void	tableEnd(void);
static void	caption(const String, calign_t);
static void	trStart(halign_t, valign_t);
static void	trEnd(void);
static void	thStart(Attrib *);
static void	thEnd(void);
static void	tdStart(Attrib *);
static void	tdEnd(void);
static void	cell(const String, Attrib *);
static void	Print(const String);

static void	initBodyAttr(Attrib *);
static void	initUlAttr(Attrib *);
static void	initOlAttr(Attrib *);
static void	initHrAttr(Attrib *);
static void	initTableAttr(Attrib *);
static void	initCellAttr(Attrib *);

static void	printHorizontalAlign(halign_t);
static void	printVerticalAlign(valign_t);
static void	printImageAlign(align_t);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static HTML *	html32 = NULL;	/* html32 object */

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	newHTML32
 * DESCRIPTION:		create new HTML32 object
 * ARGUMENTS:		none
 * RETURN VALUE:	Successful: HTML32 object, otherwise: NULL
 */
HTML *
newHTML32(void)
{
  html32 = (HTML *)calloc(1, sizeof(HTML));
  if (html32 == NULL) {
    return NULL;
  }

  html32->html.start_html	= (void *)(htmlStart);
  html32->html.end_html	= (void *)(htmlEnd);
  /**/
  html32->html.body.start	= (void *)(bodyStart);
  html32->html.body.end		= (void *)(bodyEnd);
  html32->html.body.clear	= (void *)(initBodyAttr);
  /**/
  html32->html.h1		= (void *)(h1);
  html32->html.h2		= (void *)(h2);
  html32->html.h3		= (void *)(h3);
  html32->html.h4		= (void *)(h4);
  html32->html.h5		= (void *)(h5);
  html32->html.h6		= (void *)(h6);
  /**/
  html32->html.address.begin	= (void *)(addressBegin);
  html32->html.address.end	= (void *)(addressEnd);
  html32->html.pre.begin	= (void *)(preBegin);
  html32->html.pre.end		= (void *)(preEnd);
  html32->html.center.begin	= (void *)(centerBegin);
  html32->html.center.end	= (void *)(centerEnd);
  /**/
  html32->html.p.start		= (void *)(pBegin);
  html32->html.p.end		= (void *)(pEnd);
  html32->html.div.start	= (void *)(divBegin);
  html32->html.div.end		= (void *)(divEnd);
  /**/
  html32->html.ul.start		= (void *)(ulBegin);
  html32->html.ul.end		= (void *)(ulEnd);
  html32->html.ul.clear		= (void *)(initUlAttr);
  html32->html.ol.start		= (void *)(olBegin);
  html32->html.ol.end		= (void *)(olEnd);
  html32->html.ol.clear		= (void *)(initOlAttr);
  html32->html.li		= (void *)(li);
  /**/
  html32->html.a.href		= (void *)(aHref);
  html32->html.a.name		= (void *)(aName);
  /**/
  html32->html.strong		= (void *)(strong);
  html32->html.font		= (void *)(font);
  /**/
  html32->html.br		= (void *)(br);
  /**/
  html32->html.hr.start		= (void *)(hr);
  html32->html.hr.clear		= (void *)(initHrAttr);
  /**/
  html32->html.img		= (void *)(img);
  /**/
  html32->html.form.start	= (void *)(formStart);
  html32->html.form.end		= (void *)(formEnd);
  html32->html.multipart_form.start	= (void *)(multipartStart);
  html32->html.multipart_form.end	= (void *)(formEnd);
  html32->html.input.text	= (void *)(inputText);
  html32->html.input.password	= (void *)(inputPassword);
  html32->html.input.checkbox	= (void *)(inputCheckbox);
  html32->html.input.radio	= (void *)(inputRadio);
  html32->html.input.submit	= (void *)(inputSubmit);
  html32->html.input.reset	= (void *)(inputReset);
  html32->html.input.hidden	= (void *)(inputHidden);
  html32->html.input.image	= (void *)(inputImage);
  html32->html.input.file	= (void *)(inputFile);
  html32->html.select.start	= (void *)(selectStart);
  html32->html.select.end	= (void *)(selectEnd);
  html32->html.select.option	= (void *)(selectOption);
  html32->html.textarea.start	= (void *)(textareaStart);
  html32->html.textarea.end	= (void *)(textareaEnd);
  /**/
  html32->html.table.start	= (void *)(tableStart);
  html32->html.table.end	= (void *)(tableEnd);
  html32->html.table.clear	= (void *)(initTableAttr);
  html32->html.table.caption	= (void *)(caption);
  /**/
  html32->html.tr.start		= (void *)(trStart);
  html32->html.tr.end		= (void *)(trEnd);
  html32->html.th.start		= (void *)(thStart);
  html32->html.th.end		= (void *)(thEnd);
  html32->html.th.clear		= (void *)(initCellAttr);
  html32->html.td.start		= (void *)(tdStart);
  html32->html.td.end		= (void *)(tdEnd);
  html32->html.td.clear		= (void *)(initCellAttr);
  /**/
  html32->html.print		= (void *)(Print);
  html32->html.printf		= (int *)(printf);
  /**/
  html32->html.done		= (void *)(html32Done);

  return html32;
}

static void
html32Done(void)
{
  if (html32) {
    free(html32);
    html32 = NULL;
  }
}

/*****************************************************************************/

static void
htmlStart(title)
	const String	title;
{
  fputs("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">\n"
	"<html>\n"
	"<head><title>", stdout);
  fprintf(stdout, "%s", title == NULL ? "No title" : title);
  fputs("</title></head>\n", stdout);
}

static void
htmlEnd(void)
{
  fputs("</html>\n", stdout);
}

static void
bodyStart(attr)
	register Attrib *	attr;
{
  fputs("<body", stdout);
  if (attr != NULL) {
    if (attr->x.body.bgcolor != NULL) {
      fprintf(stdout, " bgcolor=\"%s\"", attr->x.body.bgcolor);
    }
    if (attr->x.body.text != NULL) {
      fprintf(stdout, " text=\"%s\"", attr->x.body.text);
    }
    if (attr->x.body.link != NULL) {
      fprintf(stdout, " link=\"%s\"", attr->x.body.link);
    }
    if (attr->x.body.vlink != NULL) {
      fprintf(stdout, " vlink=\"%s\"", attr->x.body.vlink);
    }
    if (attr->x.body.alink != NULL) {
      fprintf(stdout, " alink=\"%s\"", attr->x.body.alink);
    }
    if (attr->x.body.background != NULL) {
      fprintf(stdout, " background=\"%s\"", attr->x.body.background);
    }
  }
  fputs(">\n", stdout);
}

static void
bodyEnd(void)
{
  fputs("</body>\n", stdout);
}

/**/
static void
h1(text, halign)
	const String	text;
	halign_t	halign;
{
  headings(1, text, halign);
}

static void
h2(text, halign)
	const String	text;
	halign_t	halign;
{
  headings(2, text, halign);
}

static void
h3(text, halign)
	const String	text;
	halign_t	halign;
{
  headings(3, text, halign);
}

static void
h4(text, halign)
	const String	text;
	halign_t	halign;
{
  headings(4, text, halign);
}

static void
h5(text, halign)
	const String	text;
	halign_t	halign;
{
  headings(5, text, halign);
}

static void
h6(text, halign)
	const String	text;
	halign_t	halign;
{
  headings(6, text, halign);
}

static void
headings(n, text, halign)
	int		n;
	const String	text;
	halign_t	halign;
{
  fprintf(stdout, "<h%d", n);
  printHorizontalAlign(halign);
  fprintf(stdout, ">%s</h%d>\n", text, n);
}

/**/

static void addressBegin(void)		{ block_begin("address", HALIGN_NONE); }
static void addressEnd(void)		{ block_end("address"); }
static void preBegin(void)		{ block_begin("pre", HALIGN_NONE); }
static void preEnd(void)		{ block_end("pre"); }
static void centerBegin(void)		{ block_begin("center", HALIGN_NONE); }
static void centerEnd(void)		{ block_end("center"); }
static void pBegin(halign_t halign)	{ block_begin("p", halign); }
static void pEnd(void)			{ block_end("p"); }
static void divBegin(halign_t halign)	{ block_begin("div", halign); }
static void divEnd(void)		{ block_end("div"); }

static void
block_begin(tag, halign)
	const String	tag;
	halign_t	halign;
{
  fprintf(stdout, "<%s", tag);
  printHorizontalAlign(halign);
  fputs(">\n", stdout);
}

static void
block_end(tag)
	const String	tag;
{
  fprintf(stdout, "</%s>\n", tag);
}

/**/

static void
ulBegin(attr)
	register Attrib *	attr;
{
  fputs("<ul", stdout);
  if (attr  != NULL) {
    switch (attr->x.ul.bullet)  {
      case BULLET_DISC:
	fputs(" bullet=\"disc\"", stdout);
	break;
      case BULLET_SQUARE:
	fputs(" bullet=\"square\"", stdout);
	break;
      case BULLET_CIRCLE:
	fputs(" bullet=\"circle\"", stdout);
	break;
      default:
	break;
    }
    if (attr->x.ul.compact) {
      fputs(" compact", stdout);
    }
  }
  fputs(">\n", stdout);
}

static void
ulEnd(void)
{
  fputs("</ul>\n", stdout);
}

static void
olBegin(attr)
	register Attrib *	attr;
{
  fputs("<ol", stdout);
  if (attr  != NULL) {
    switch (attr->x.ol.numbering)  {
      case NUMBERING_ARABIC_NUMBER:
	fputs(" numbering=\"1\"", stdout);
	break;
      case NUMBERING_LOWER_ALPHA:
	fputs(" numbering=\"a\"", stdout);
	break;
      case NUMBERING_UPPER_ALPHA:
	fputs(" numbering=\"A\"", stdout);
	break;
      case NUMBERING_LOWER_ROMAN:
	fputs(" numbering=\"i\"", stdout);
	break;
      case NUMBERING_UPPER_ROMAN:
	fputs(" numbering=\"I\"", stdout);
	break;
      default:
	break;
    }
    if (attr->x.ol.start > 0) {
      fprintf(stdout, " start=\"%d\"", attr->x.ol.start);
    }
    if (attr->x.ol.compact) {
      fputs(" compact", stdout);
    }
  }
  fputs(">\n", stdout);
}

static void
olEnd(void)
{
  fputs("</ol>\n", stdout);
}

static void
li(text)
	const String	text;
{
  if (text != NULL) {
    fprintf(stdout, "<li>%s\n", text);
  }
}

/**/

static void
aHref(uri, text)
	const String	uri;
	const String	text;
{
  fprintf(stdout, "<a href=\"%s\">%s</a>", uri, text);
}

static void
aName(name, text)
	const String	name;
	const String	text;
{
  fprintf(stdout, "<a name=\"%s\">%s</a>", name, text);
}

/**/

static void
strong(text)
	const String	text;
{
  fprintf(stdout, "<strong>%s</strong>", text);
}

static void
font(text, size, color)
	const String	text;
	const String	size;
	const String	color;
{
  fputs("<font", stdout);
  if (size != NULL) {
    fprintf(stdout, " size=\"%s\"", size);
  }
  if (color != NULL) {
    fprintf(stdout, " color=\"%s\"", color);
  }
  fprintf(stdout, ">%s</font>", text);
}

/**/

static void br(void)	{ fputs("<br>\n", stdout); }

static void
hr(attr)
	register Attrib *	attr;
{
  fputs("<hr", stdout);
  if (attr != NULL) {
    printHorizontalAlign(attr->x.hr.halign);
    if (attr->x.hr.size > 0) {
      fprintf(stdout, " size=\"%d\"", attr->x.hr.size);
    }
    if (attr->x.hr.width != NULL) {
      fprintf(stdout, " width=\"%s\"", attr->x.hr.width);
    }
    if (attr->x.hr.noshade) {
      fputs(" noshade", stdout);
    }
  }
  fputs(">\n", stdout);
}

/**/

static void
img(src, alt, width, height, border, align)
	const String	src;
	const String	alt;
	Pixels		width;
	Pixels		height;
	Pixels		border;
	align_t		align;
{
  fprintf(stdout, "<img src=\"%s\" alt=\"%s\" width=\"%d\" height=\"%d\"",
	  src, alt, width, height);
  if (border >= 0) {
    fprintf(stdout, " border=\"%d\"", border);
  }
  printImageAlign(align);
  fputs(">\n", stdout);
}

/**/

static void
formStart(action, method)
	const String	action;
	method_t	method;
{
  fprintf(stdout, "<form action=\"%s\" method=", action);
  switch (method) {
    case METHOD_POST:
      fputs("\"post\"", stdout);
      break;
    case METHOD_GET:
    default:
      fputs("\"get\"", stdout);
      break;
  }
  fputs(" enctype=\"application/x-www-form-urlencoded\">\n", stdout);
}

static void
multipartStart(action, method)
	const String	action;
	method_t	method;
{
  fprintf(stdout, "<form action=\"%s\" method=", action);
  switch (method) {
    case METHOD_POST:
      fputs("\"post\"", stdout);
      break;
    case METHOD_GET:
    default:
      fputs("\"get\"", stdout);
      break;
  }
  fputs(" enctype=\"multipart/form-data\">\n", stdout);
}

static void
formEnd(void)
{
  fputs("</form>\n", stdout);
}

static void
inputText(name, value, size, maxlength)
	const String	name;
	const String	value;
	unsigned int	size;
	unsigned int	maxlength;
{
  fprintf(stdout, "<input type=\"text\" name=\"%s\" value=\"%s\" "
		  "size=\"%u\" maxlength=\"%u\">",
		  name, value, size, maxlength);
}

static void
inputPassword(name, value, size, maxlength)
	const String	name;
	const String	value;
	unsigned int	size;
	unsigned int	maxlength;
{
  fprintf(stdout, "<input type=\"password\" name=\"%s\" value=\"%s\" "
		  "size=\"%u\" maxlength=\"%u\">",
		  name, value, size, maxlength);
}

static void
inputCheckbox(name, value, checked)
	const String	name;
	const String	value;
	bool		checked;
{
  fprintf(stdout, "<input type=\"checkbox\" name=\"%s\" value=\"%s\"%s>",
		  name, value, checked ? " checked" : "");
}

static void
inputRadio(name, value, checked)
	const String	name;
	const String	value;
	bool		checked;
{
  fprintf(stdout, "<input type=\"radio\" name=\"%s\" value=\"%s\"%s>",
		  name, value, checked ? " checked" : "");
}

static void
inputSubmit(name, value)
	const String	name;
	const String	value;
{
  fprintf(stdout,
	  "<input type=\"submit\" name=\"%s\" value=\"%s\">", name, value);
}

static void
inputReset(name, value)
	const String	name;
	const String	value;
{
  fprintf(stdout,
	  "<input type=\"reset\" name=\"%s\" value=\"%s\">", name, value);
}

static void
inputHidden(name, value)
	const String	name;
	const String	value;
{
  fprintf(stdout,
	  "<input type=\"hidden\" name=\"%s\" value=\"%s\">", name, value);
}

static void
inputImage(name, src, align)
	const String	name;
	const String	src;
	align_t		align;
{
  fprintf(stdout, "<input type=\"image\" name=\"%s\" src=\"%s\"", name, src);
  printImageAlign(align);
  fputs(">\n", stdout);
}

static void
inputFile(name, value, size, maxlength)
	const String	name;
	const String	value;
	unsigned int	size;
	unsigned int	maxlength;
{
  fprintf(stdout, "<input type=\"file\" name=\"%s\" value=\"%s\" "
		  "size=\"%u\" maxlength=\"%u\">",
		  name, value, size, maxlength);
}

static void
selectStart(name, size, multiple)
	const String	name;
	unsigned int	size;
	bool		multiple;
{
  fprintf(stdout, "<select name=\"%s\"", name);
  if (size > 0) {
    fprintf(stdout, " size=\"%u\"", size);
  }
  if (multiple) {
    fputs(" multiple>\n", stdout);
  }
  fputs(">\n", stdout);
}

static void
selectEnd(void)
{
  fputs("</select>\n", stdout);
}

static void
selectOption(value, selected)
	const String	value;
	bool		selected;
{
  fprintf(stdout, "<option value=\"%s\"", value);
  if (selected) {
    fputs(" selected>\n", stdout);
  }
  fputs(">\n", stdout);
}

static void
textareaStart(name, rows, cols)
	const String	name;
	unsigned int	rows;
	unsigned int	cols;
{
  fprintf(stdout,
	  "<textarea name=\"%s\" rows=\"%u\" cols=\"%u\">\n", name, rows, cols);
}

static void
textareaEnd(void)
{
  fputs("</textarea>\n", stdout);
}

/**/

static void
tableStart(attr)
	register Attrib *	attr;
{
  fputs("<table", stdout);
  if (attr != NULL) {
    if (attr->x.table.border >= 0) {
      fprintf(stdout, " border=\"%d\"", attr->x.table.border);
    }
    if (attr->x.table.width != NULL) {
      fprintf(stdout, " width=\"%s\"", attr->x.table.width);
    }
    if (attr->x.table.cellspacing >= 0) {
      fprintf(stdout, " cellspacing=\"%d\"", attr->x.table.cellspacing);
    }
    if (attr->x.table.cellpadding >= 0) {
      fprintf(stdout, " cellpadding=\"%d\"", attr->x.table.cellpadding);
    }
    printHorizontalAlign(attr->x.table.halign);
  }
  fputs(">\n", stdout);
}

static void
tableEnd(void)
{
  fputs("</table>\n", stdout);
}

static void
caption(text, calign)
	const String	text;
	calign_t	calign;
{
  fputs("<caption", stdout);
  switch (calign) {
    case CALIGN_TOP:
      fputs(" align=\"top\"", stdout);
      break;
    case CALIGN_BOTTOM:
      fputs(" align=\"bottom\"", stdout);
      break;
    default:
      break;
  }
  fprintf(stdout, ">%s</caption>\n", text);
}

static void
trStart(halign, valign)
	halign_t	halign;
	valign_t	valign;
{
  fputs("<tr", stdout);
  printHorizontalAlign(halign);
  printVerticalAlign(valign);
  fputs(">", stdout);
}

static void
trEnd(void)
{
  fputs("</tr>\n", stdout);
}

static void
thStart(attr)
	register Attrib *	attr;
{
  cell("th", attr);
}

static void
thEnd(void)
{
  fputs("</th>", stdout);
}

static void
tdStart(attr)
	register Attrib *	attr;
{
  cell("td", attr);
}

static void
tdEnd(void)
{
  fputs("</td>", stdout);
}

static void
cell(tag, attr)
	const String		tag;
	register Attrib *	attr;
{
  fprintf(stdout, "<%s", tag);
  if (attr != NULL) {
    if (attr->x.cell.nowrap) {
      fputs(" nowrap", stdout);
    }
    if (attr->x.cell.rowspan > 0) {
      fprintf(stdout, " rowspan=\"%d\"", attr->x.cell.rowspan);
    }
    if (attr->x.cell.colspan > 0) {
      fprintf(stdout, " colspan=\"%d\"", attr->x.cell.colspan);
    }
    printHorizontalAlign(attr->x.cell.halign);
    printVerticalAlign(attr->x.cell.valign);
    if (attr->x.cell.width > 0) {
      fprintf(stdout, " width=\"%d\"", attr->x.cell.width);
    }
    if (attr->x.cell.height > 0) {
      fprintf(stdout, " height=\"%d\"", attr->x.cell.height);
    }
  }
  fputs(">", stdout);
}

static void
Print(text)
	const String	text;
{
  if (text != NULL) {
    fputs(text, stdout);
  }
}

/*****************************************************************************/

static void
initBodyAttr(attr)
	register Attrib *	attr;
{
  attr->x.body.bgcolor	  = NULL;
  attr->x.body.text	  = NULL;
  attr->x.body.link	  = NULL;
  attr->x.body.vlink	  = NULL;
  attr->x.body.alink	  = NULL;
  attr->x.body.background = NULL;
}

static void
initUlAttr(attr)
	register Attrib *	attr;
{
  attr->x.ul.bullet	= BULLET_NONE;
  attr->x.ul.compact	= false;
}

static void
initOlAttr(attr)
	register Attrib *	attr;
{
  attr->x.ol.numbering	= NUMBERING_NONE;
  attr->x.ol.start	= 1;
  attr->x.ol.compact	= false;
}

static void
initHrAttr(attr)
	register Attrib *	attr;
{
  attr->x.hr.halign	= HALIGN_NONE;
  attr->x.hr.size	= 0;
  attr->x.hr.width	= NULL;
  attr->x.hr.noshade	= false;
}

static void
initTableAttr(attr)
	register Attrib *	attr;
{
  attr->x.table.border		= 1;
  attr->x.table.width		= NULL;
  attr->x.table.cellspacing	= 0;
  attr->x.table.cellpadding	= 0;
  attr->x.table.halign		= HALIGN_NONE;
}

static void
initCellAttr(attr)
	register Attrib *	attr;
{
  attr->x.cell.nowrap	= false;
  attr->x.cell.rowspan	= 0;
  attr->x.cell.colspan	= 0;
  attr->x.cell.halign	= HALIGN_NONE;
  attr->x.cell.valign	= VALIGN_NONE;
  attr->x.cell.width	= 0;
  attr->x.cell.height	= 0;
}

/*****************************************************************************/

static void
printHorizontalAlign(halign)
	halign_t	halign;
{
  switch (halign) {
    case HALIGN_LEFT:
      fputs(" align=\"left\"", stdout);
      break;
    case HALIGN_CENTER:
      fputs(" align=\"center\"", stdout);
      break;
    case HALIGN_RIGHT:
      fputs(" align=\"right\"", stdout);
      break;
    default:
      break;
  }
}

static void
printVerticalAlign(valign)
	valign_t	valign;
{
  switch (valign) {
    case VALIGN_TOP:
      fputs(" valign=\"top\"", stdout);
      break;
    case VALIGN_MIDDLE:
      fputs(" valign=\"middle\"", stdout);
      break;
    case VALIGN_BOTTOM:
      fputs(" valign=\"bottom\"", stdout);
      break;
    default:
      break;
  }
}

static void
printImageAlign(align)
	align_t	align;
{
  switch (align) {
    case ALIGN_TOP:
      fputs(" align=\"top\"", stdout);
      break;
    case ALIGN_MIDDLE:
      fputs(" align=\"middle\"", stdout);
      break;
    case ALIGN_BOTTOM:
      fputs(" align=\"bottom\"", stdout);
      break;
    case ALIGN_LEFT:
      fputs(" align=\"left\"", stdout);
      break;
    case ALIGN_RIGHT:
      fputs(" align=\"right\"", stdout);
      break;
    default:
      break;
  }
}
