/*****************************************************************************
 *
 * FILE:	mod_pgsql.c
 * DESCRIPTION:	Cockatrice: PostgreSQL database operator (module)
 * DATE:	Sat, Aug 18 2007
 * UPDATED:	Sun, Aug 26 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_pgsql.c,v 1.3 2007/08/26 12:18:06 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <libpq-fe.h>
#include "cockatrice.h"
#include "module.h"
#include "ccpath.h"
#include "functions.h"

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	cm_pgsql_finish(void);
static int	cm_pgsql_open(const char *);
static int	cm_pgsql_close(void);
static int	cm_pgsql_exec(const char *);
static int	cm_pgsql_query(const char *);
static int	cm_pgsql_fetch(const char ***);
static char *	cm_pgsql_quote(const char *);
static int	escape(const char * s, char ** r);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static PGconn *		conn = NULL;
static PGresult *	res  = NULL;
static int		max  = 0;	/* value of PQntuples(); */
static int		num  = 0;	/* value of PQnfields(); */
static int		cur  = 0;	/* current tuple */

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

static void
cm_pgsql_finish(void)
{
  cm_pgsql_close();
}

/*
 * FUNCTION NAME:	cm_pgsql_open
 * DESCRIPTION:		connect the database server
 * ARGUMENTS:		connect info
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_pgsql_open(conninfo)
	const char *	conninfo;
{
  if (conn != NULL) {
    PQfinish(conn);
  }

  conn = PQconnectdb(conninfo);

  if (PQstatus(conn) == CONNECTION_BAD) {
    fprintf(stderr, "Failed to connect database: %s\n", PQerrorMessage(conn));
    PQfinish(conn);
    cgi_errno = CGIERR_DATABASE_OPEN_FAILURE;
    return -1;
  }

  return 0;
}

/*
 * FUNCTION NAME:	cm_pgsql_close
 * DESCRIPTION:		disconnect the database server
 * ARGUMENTS:		none
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_pgsql_close(void)
{
  if (res != NULL) {
    PQclear(res);
    res = NULL;
  }

  if (conn != NULL) {
    PQfinish(conn);
    conn = NULL;
  }

  return 0;
}

/*
 * FUNCTION NAME:	cm_pgsql_exec
 * DESCRIPTION:		exec a sql statement
 * ARGUMENTS:		SQL statement
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_pgsql_exec(sql)
	const char *	sql;
{
  if (sql  == NULL) { return -1; }	/* no SQL query */
  if (conn == NULL) { return -1; }	/* no opened database */

  if (res != NULL) {
    PQclear(res);
    res = NULL;
  }

  if (PQstatus(conn) != CONNECTION_OK) {
    PQreset(conn);
  }
  res = PQexec(conn, sql);

  switch (PQresultStatus(res)) {
    case PGRES_TUPLES_OK:
      max = PQntuples(res);
      num = PQnfields(res);
      cur = 0;
    case PGRES_COMMAND_OK:
      return 0;
    case PGRES_COPY_OUT:
    case PGRES_COPY_IN:
      return 1;
    case PGRES_EMPTY_QUERY:
    case PGRES_BAD_RESPONSE:
    case PGRES_NONFATAL_ERROR:
    case PGRES_FATAL_ERROR:
    default:
      fprintf(stderr, "Failed: %s", PQresultErrorMessage(res));
      break;
  }

  return -1;
}

/*
 * FUNCTION NAME:	cm_pgsql_query
 * DESCRIPTION:		send a SQL query
 * ARGUMENTS:		SQL statement
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_pgsql_query(sql)
	const char *	sql;
{
  return cm_pgsql_exec(sql);
}

/*
 * FUNCTION NAME:	cm_pgsql_fetch
 * DESCRIPTION:		fetch a result
 * ARGUMENTS:		pointer for result
 * RETURN VALUE:	done: 0, error: -1, column of result: >0
 */
static int
cm_pgsql_fetch(retval)
	const char ***	retval;
{
  if (conn == NULL) { return -1; }	/* no opened database */
  if (res != NULL && cur < max) {
    const char **	value;

    value = (const char **)cc_calloc((size_t)num, sizeof(const char *));
    if (value != NULL) {
      register int	i;

      for (i = 0; i < num; i++) {
	char *	v;

	v = PQgetvalue(res, cur, i);
	if (v != NULL) {
	  value[i] = (char *)cc_strdup(v);
	}
      }
      cur++;
      if (retval != NULL) {
	*retval = value;
	return num;
      }
    }
  }

  return -1;
}

/*
 * FUNCTION NAME:	cm_pgsql_quote
 * DESCRIPTION:		quote a SQL statement
 * ARGUMENTS:		SQL statement
 * RETURN VALUE:	successful: a quoted SQL statement, otherwise: NULL
 */
static char *
cm_pgsql_quote(sql)
	const char *	sql;
{
  char *	ret;

  return (escape(sql, &ret) == 0) ? ret : NULL;
}

static int
escape(s, r)
	const char *	s;
	char **		r;	/* return value */
{
  int	error = -1;

  if (s != NULL) {
    size_t	l;

    l  = strlen(s);
    *r = (char *)malloc(l * 2 + 1);
    if (*r != NULL) {
      PQescapeStringConn(conn, *r, s, l, &error);
      if (error != 0) {	/* failed */
	free(*r);
	*r = NULL;
      }
    }
  }

  return error;
}

/*****************************************************************************/

static const cmDatabase	pgsql_funcs = {
  cm_pgsql_open,
  cm_pgsql_close,
  cm_pgsql_exec,
  cm_pgsql_query,
  cm_pgsql_fetch,
  cm_pgsql_quote
};

static void
cm_pgsql_register_hooks(void)
{
  cc_hook_module_database(&pgsql_funcs);
}

static ccModule	mod_pgsql = {
  CC_MODULE_MAGIC_NUMBER_MAJOR,		/* API version */
  "mod_pgsql",				/* module name */
  "Cockatrice: PostgreSQL module 1.0.1",	/* description such as version etc.. */
  NULL,					/* module handle */
  (void *)NULL,				/* called when it's loaded */
  (void *)cm_pgsql_finish,		/* called when it's unloaded */
  (void *)cm_pgsql_register_hooks
};

ccModule *
ccModuleInfo(void)
{
  return &mod_pgsql;
}
